/* POWER MANAGEMENT ACPI DAEMON
 * Copyright (C) 2000  Monta Vista Software Inc.
 */
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <sys/file.h>
#include "pm.h"
#include "batt.h"

#ifdef PROTOTYPE
#define ACPI_EVENT_DEV	"/proc/sys/acpi/event"
#if 0
#define ACPI_EVENT_DEV	"event"
#endif
#define ACPI_BATT_DEV	"battinfo"
#else
#define ACPI_EVENT_DEV	"/proc/sys/acpi/event"
#define ACPI_BATT_DEV	"/proc/sys/acpi/power_source/...."
#endif

/* time interval for polling AC and battery status on sm-bus
 */
#define BATT_POLL_TIME	60	/* slow but safe */

/* PM1_STS/EN flags [from: linux-2.3.99-pre8/include/linux/acpi.h]
 */
#define ACPI_TMR    0x0001
#define ACPI_BM     0x0010
#define ACPI_GBL    0x0020
#define ACPI_PWRBTN 0x0100
#define ACPI_SLPBTN 0x0200
#define ACPI_RTC    0x0400
#define ACPI_WAK    0x8000

typedef struct {
	char *param;
	int val;
	} RTP;

RTP rtp_curr[] = {
	{"battery_level", 4},
	{"battery_present", 0},
	{"powerbutton_pressed", 0},
	{"ac_input", 1},
	{"charging", 1},
	{NULL, 0}};

int child_pid = {0};
int pending_pb = {0};
char *procname;

RTP *rtp_get(char *param)
	{
	RTP *rtp;

	for (rtp = rtp_curr; rtp->param; ++rtp)
		if (!strcmp(rtp->param, param))
			return (rtp);
	return (NULL);
	}

/* update paramenter, indicate if val has changed
 */
int rtp_set(char *param, int val)
	{
	RTP *rtp;

	if (!(rtp = rtp_get(param)))
		;
	else if (rtp->val == val)
		;
	else
		{
		rtp->val = val;
		return (1);
		}
	return (0);
	}
			
/* wack the pm daemon when the power status file changes
 */
char *signal_daemon(char *fn, int sig)
	{
	FILE *file;
	unsigned int pid;

	if (!(file = fopen(fn, "r")))
		return ("can't open daemon pid file");
	else if (fscanf(file, "%d", &pid) != 1)
		return ("corrupt daemon pid file");
	else if (kill(pid, sig) < 0)
		return ("can't signal daemon");
	return (NULL);
	}

/* read in all parameters before modification
 * NB: write lock asumed held on the target file
 */
char *read_params()
	{
	char name[100];	/* see fscanf %s arg below */
	FILE *file;
	RTP *p;
	int val;

	if (!(file = fopen(POWER_STATUS_PATH, "r")))
		perror("can't re-open file");
	else
		while (fscanf(file, "%99s %d", name, &val) == 2)
			if (!(p = rtp_get(name)))
				fprintf(stderr, "Warning unknown parameter %s\n", name);
			else
				p->val = val;
	fclose(file);
	}

/* generate power status file
 */
int output_status()
	{
	static int err_cnt = {0};
	FILE *file;
	char *err, *sigerr;
	RTP *rtp;

	err = NULL;
	if (!(file = fopen(POWER_STATUS_PATH, "w")))
		err = "open";
	else if (flock(file->_fileno, LOCK_EX|LOCK_NB))
		err = "lock";
	else
		{
		for (rtp = rtp_curr; rtp->param; ++rtp)
			fprintf(file, "%s %d\n", rtp->param, rtp->val);
		}
	if (file)
		{
		flock(file->_fileno, LOCK_UN);
		fclose(file);
		}
	if (!err)
		{
		err_cnt = 0;
		if (sigerr = signal_daemon(PMD_PID_PATH, SIGUSR1))
			perror(sigerr);
		}
	else if (++err_cnt < 3)
		;
	else
		fprintf(stderr, "Daemon status file error: %s\n", err);
	return (1);
	}

void cleanup(int sig)
	{
	if (child_pid && kill(child_pid, SIGINT) < 0)
		perror("couldn't kill child");
	_exit(0);
	}

/* SIGUSR1 handler
 */
void pb_pressed()
	{
 	++rtp_get("powerbutton_pressed")->val;
	output_status();
	}

/* a process whose sole purpose in life is to read /proc/sys/event
 * and signal the pmd when the power button is pressed.  Very ugly
 * indeed but the current arch/i386/kernel/acpi.c driver doesn't
 * support non-blocking reads or signal generation.  Quite
 * honestly the ACPI driver, second only to the God Forsaken ACPI
 * "mechanism", is one of the ugliest, tortured creations invented
 * by man.
 */
int main(int argc, char **argv)
	{
	extern BINFO *get_batt_info(BINFO *);
	static BINFO prev_binfo;
	unsigned int pm1_status, gpe_status, event_state;
	FILE *file;
	char *err;
	BINFO binfo;
	int mask;

	procname = *argv;
	output_status();
	if (!(child_pid = fork()))
		{	/* child */
		signal(SIGUSR1, pb_pressed);
#if 0
		if (!(file = fopen(ACPI_BATT_DEV, "r")))
			{
			perror("Can't open: "ACPI_BATT_DEV);
			exit(5);
			}
#endif
		for (;;)
			{
			if (get_batt_data(&binfo), binfo.s != prev_binfo.s)
				{
				mask = sigblock(sigmask(SIGUSR1));
				rtp_set("battery_present", binfo.b.present);
				rtp_set("battery_level", binfo.b.capacity);
				rtp_set("ac_input", binfo.b.ac);
				rtp_set("charging", binfo.b.charging);
				output_status();
				sigsetmask(mask);
				prev_binfo = binfo;
				}
			sleep(BATT_POLL_TIME);
			}
		}
	else if (child_pid < 0)
		{
		perror("can't fork child");
		exit(1);
		}
	else
		{	/* parent */
		signal(SIGINT, cleanup);
		on_exit(cleanup, NULL);
		if (!(file = fopen(ACPI_EVENT_DEV, "r")))
			{
			perror("Can't open: "ACPI_EVENT_DEV);
			exit(5);
			}
		for (;;)
			if (fscanf(file, "%i %i %i",
				&pm1_status, &gpe_status, &event_state) != 3)
					;
			else if (pm1_status & ACPI_PWRBTN)
				kill(child_pid, SIGUSR1);
		}
	}
